/*
*	aritmetiko.c - библиотека арифметических функций.
*	© 2005, А.А.Владимиров.
*
*	Данная программа является свободным программным обеспечением.
*	Вы вправе распространять ее и/или модифицировать в соответствии
*	с условиями версии 2, либо, по вашему выбору, с условиями
*	более поздней версии Стандартной Общественной Лицензии GNU,
*	опубликованной Free Software Foundation.
*
*	Мы распространяем данную программу в надежде на то, что она будет вам
*	полезной, однако НЕ ПРЕДОСТАВЛЯЕМ НА НЕЕ НИКАКИХ ГАРАНТИЙ, в том числе
*	ГАРАНТИИ ТОВАРНОГО СОСТОЯНИЯ ПРИ ПРОДАЖЕ и ПРИГОДНОСТИ ДЛЯ ИСПОЛЬЗОВАНИЯ
*	В КОНКРЕТНЫХ ЦЕЛЯХ. Для получения более подробной информации ознакомьтесь
*	со Стандартной Общественной Лицензией GNU.
*
*	Вместе с данной программой вы должны были получить экземпляр
*	Стандартной Общественной Лицензии GNU. Если вы его не получили,
*	сообщите об этом в
*	
*	Free Software Foundation, Inc.,
*	59 Temple Place, Suite 330,
*	Boston, MA 02111-1307
*	USA
*
*/

# include "refal.h"


void arit_sxargi (void) __attribute__ ((constructor));


void arit_add (void);
void arit_sub (void);
void arit_mul (void);
void arit_divmod (void);
void arit_div (void);
void arit_mod (void);


void arit_sumigi (LITERO*, unsigned long);
void arit_subtrahi (LITERO*, unsigned long);


/* ========================================================================== */


extern void ref_eraro (int);
extern void interp_aldoni_enkonstruitan (char* nomo, void (*funk) (void));


/* ========================================================================== */


void arit_sxargi () {
	
	interp_aldoni_enkonstruitan ("add", arit_add);
	interp_aldoni_enkonstruitan ("sub", arit_sub);
	interp_aldoni_enkonstruitan ("mul", arit_mul);
	interp_aldoni_enkonstruitan ("divmod", arit_divmod);
	interp_aldoni_enkonstruitan ("div", arit_div);
	interp_aldoni_enkonstruitan ("mod", arit_mod);
	return;
}


/* ========================================================================== */


/*
*	Функция "arit_add" производит сложение пары чисел, заключённых между
*	KKK и KKP.
*
*/

void arit_add () {
	LITERO *kuranta1, *kuranta2;
	unsigned long k = 0;
	
	kuranta1 = KKK->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta1)) ref_eraro (8);
	kuranta2 = (kuranta1->dat.dual)->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta2)) ref_eraro (8);
	if ((kuranta2->dat.dual)->sekv != KKP) ref_eraro (8);
	kuranta1 = kuranta1->sekv;
	kuranta2 = kuranta2->sekv;
	detrui_literon (kuranta2->ant);
	while (ESTAS_CIFER (kuranta1) && ESTAS_CIFER (kuranta2)) {
		k += (unsigned long) kuranta1->dat.cif + 
			(unsigned long) kuranta2->dat.cif;
		kuranta1->dat.cif = (unsigned int) (k % 65536);
		k /= 65536;
		kuranta1 = kuranta1->sekv;
		kuranta2 = kuranta2->sekv;
		detrui_literon (kuranta2->ant);
	};
	if (ESTAS_D_LIT_PARENT (kuranta2)) {
		while (ESTAS_CIFER (kuranta1)) {
			k += (unsigned long) kuranta1->dat.cif;
			kuranta1->dat.cif = (unsigned int) (k % 65536);
			k /= 65536;
			kuranta1 = kuranta1->sekv;
		};
		ligi_parentezojn (kuranta1->dat.dual, kuranta2);
		if (k != 0) {
			kuranta1->tip = CIFER;
			kuranta1->dat.cif = (unsigned int) k;
			ligi_literojn (kuranta1, kuranta2);
		}
		else {
			detrui_literon (kuranta1);
			ligi_literojn (LLM->ant, kuranta2);
		};
	}
	else {
		kuranta1 = kuranta1->dat.dual;
		(kuranta1->dat.dual)->tip = CIFER;
		ligi_literojn (kuranta1->dat.dual, kuranta2);
		while (ESTAS_CIFER (kuranta2)) {
			k += (unsigned long) kuranta2->dat.cif;
			(kuranta2->ant)->dat.cif = (unsigned int) (k % 65536);
			k /= 65536;
			kuranta2 = kuranta2->sekv;
		};
		ligi_parentezojn (kuranta1, kuranta2);
		if (k != 0) {
			(kuranta2->ant)->dat.cif = (unsigned int) k;
		}
		else {
			detrui_literon (kuranta2->ant);
			ligi_literojn (LLM->ant, kuranta2);
		};
	};
	return;
}


/*
*	Функция "arit_sub" производит вычитание пары чисел, заключённых между
*	KKK и KKP.
*
*	ЗАМЕЧАНИЕ: осуществляется операция, обозначаемая иногда 'минусом с точкой':
*	результатом вычитания большего числа из меньшего по определению
*	считается 0.
*
*/

void arit_sub () {
	LITERO *kuranta1, *kuranta2;
	unsigned long k = 1;
	
	kuranta1 = KKK->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta1)) ref_eraro (8);
	kuranta2 = (kuranta1->dat.dual)->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta2)) ref_eraro (8);
	if ((kuranta2->dat.dual)->sekv != KKP) ref_eraro (8);
	kuranta1 = kuranta1->sekv;
	kuranta2 = kuranta2->sekv;
	detrui_literon (kuranta2->ant);
	while (ESTAS_CIFER (kuranta1) && ESTAS_CIFER (kuranta2)) {
		k += 65535 + (unsigned long) (kuranta1->dat.cif) -
			(unsigned long) (kuranta2->dat.cif);
		kuranta1->dat.cif = (unsigned int) (k % 65536);
		k /= 65536;
		kuranta1 = kuranta1->sekv;
		kuranta2 = kuranta2->sekv;
		detrui_literon (kuranta2->ant);
	};
	if (ESTAS_D_LIT_PARENT (kuranta2)) {
		while (ESTAS_CIFER (kuranta1) && (k != 1)) {
			k += 65535 + (unsigned long) (kuranta1->dat.cif);
			kuranta1->dat.cif = (unsigned int) (k % 65536);
			k /= 65536;
			kuranta1 = kuranta1->sekv;
		};
		if (k != 1) {
			while (!ESTAS_MD_LIT_PARENT (kuranta1->ant)) {
				detrui_literon (kuranta1);
				kuranta1 = kuranta1->ant;
			};
			ligi_literojn (kuranta1, kuranta2);
			ligi_parentezojn (kuranta1->ant, kuranta2);
			kuranta1->dat.cif = 0;
		}
		else {
			while (!ESTAS_D_LIT_PARENT (kuranta1)) {
				kuranta1 = kuranta1->sekv;
			};
			ligi_literojn (kuranta1->ant, kuranta2);
			ligi_parentezojn (kuranta1->dat.dual, kuranta2);
			detrui_literon (kuranta1);
		};
	}
	else {
		while (!ESTAS_D_LIT_PARENT (kuranta2)) {
			kuranta2 = kuranta2->sekv;
			detrui_literon (kuranta2->ant);
		};
		while (!ESTAS_MD_LIT_PARENT (kuranta1->ant)) {
			detrui_literon (kuranta1);
			kuranta1 = kuranta1->ant;
		};
		ligi_literojn (kuranta1, kuranta2);
		ligi_parentezojn (kuranta1->ant, kuranta2);
		kuranta1->dat.cif = 0;
	};
	kuranta1 = KKK->sekv;
	kuranta2 = (KKP->ant)->ant;
	while ((kuranta2->dat.cif == 0) && (kuranta2->ant != kuranta1)) {
		ligi_literojn (kuranta2->ant, kuranta2->sekv);
		detrui_literon (kuranta2);
		kuranta2 = kuranta2->ant;
	};
	return;
}


/*
*	Функция "arit_mul" производит перемножение пары чисел, заключённых между
*	KKK и KKP.
*
*/

void arit_mul () {
	unsigned long k;
	LITERO *kuranta1, *kuranta2, *kuranta0, *kuranta;
	
	kuranta1 = KKK->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta1)) ref_eraro (8);
	kuranta2 = (kuranta1->dat.dual)->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta2)) ref_eraro (8);
	if ((kuranta2->dat.dual)->sekv != KKP) ref_eraro (8);
	kuranta1 = kuranta1->sekv;
	kuranta2 = kuranta2->sekv;
	krei_literon (KKK->sekv);
	(KKK->sekv)->ant = KKK;
	(KKK->sekv)->tip = MD_LIT_PARENT;
	krei_literon (KKP->ant);
	(KKP->ant)->sekv = KKP;
	(KKP->ant)->tip = D_LIT_PARENT;
	ligi_parentezojn (KKK->sekv, KKP->ant);
	krei_literon (kuranta);
	kuranta->tip = CIFER;
	kuranta->dat.cif = 0;
	ligi_literojn (KKK->sekv, kuranta);
	ligi_literojn (kuranta, KKP->ant);
	kuranta0 = kuranta;
	while (!ESTAS_D_LIT_PARENT (kuranta2)) {
		while (!ESTAS_D_LIT_PARENT (kuranta1)) {
			k = (unsigned long) (kuranta1->dat.cif) * 
				(unsigned long) (kuranta2->dat.cif);
			arit_sumigi (kuranta, k);
			kuranta1 = kuranta1->sekv;
			kuranta = kuranta->sekv;
			if (ESTAS_D_LIT_PARENT (kuranta)) {
				krei_literon ((kuranta->ant)->sekv);
				((kuranta->ant)->sekv)->ant = kuranta->ant;
				ligi_literojn ((kuranta->ant)->sekv, kuranta);
				kuranta = kuranta->ant;
				kuranta->tip = CIFER;
				kuranta->dat.cif = 0;
			};
		};
		kuranta2 = kuranta2->sekv;
		kuranta1 = (kuranta1->dat.dual)->sekv;
		kuranta0 = kuranta0->sekv;
		if (ESTAS_D_LIT_PARENT (kuranta0)) {
			krei_literon ((kuranta0->ant)->sekv);
			((kuranta0->ant)->sekv)->ant = kuranta0->ant;
			ligi_literojn ((kuranta0->ant)->sekv, kuranta0);
			kuranta0 = kuranta0->ant;
			kuranta0->tip = CIFER;
			kuranta0->dat.cif = 0;
		};
		kuranta = kuranta0;
	};
	do {
		detrui_literon (kuranta2);
		kuranta2 = kuranta2->ant;
	} while (!ESTAS_MD_LIT_PARENT (kuranta2));
	do {
		detrui_literon (kuranta2);
		kuranta2 = kuranta2->ant;
	} while (!ESTAS_MD_LIT_PARENT (kuranta2));
	detrui_literon (kuranta2);
	kuranta = (KKP->ant)->ant;
	while ((kuranta->dat.cif == 0) && ESTAS_CIFER (kuranta->ant)) {
		ligi_literojn (kuranta->ant, kuranta->sekv);
		detrui_literon (kuranta);
		kuranta = kuranta->ant;
	};
	return;
}


/*
*	Функция "arit_divmod" производит целочисленное деление с остатком
*	пары чисел, заключённых между KKK и KKP. Результатом является
*	пара из целочисленного частного и остатка.
*
*/

void arit_divmod (void) {
	unsigned long k;
	int i = 1;
	LITERO *kuranta1, *kuranta2, *kuranta3, *lasta1, *kuranta;
	
	kuranta1 = KKK->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta1)) ref_eraro (8);
	kuranta2 = (kuranta1->dat.dual)->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta2)) ref_eraro (8);
	if ((kuranta2->dat.dual)->sekv != KKP) ref_eraro (8);
	if (((kuranta2->dat.dual)->ant)->dat.cif == 0) ref_eraro (8);
	krei_literon (kuranta);
	kuranta->tip = MD_LIT_PARENT;
	ligi_literojn (KKK, kuranta);
	krei_literon (kuranta);
	kuranta->tip = CIFER;
	kuranta->dat.cif = 0;
	ligi_literojn (KKK->sekv, kuranta);
	krei_literon (kuranta->sekv);
	(kuranta->sekv)->tip = D_LIT_PARENT;
	(kuranta->sekv)->ant = kuranta;
	ligi_parentezojn (kuranta->ant, kuranta->sekv);
	ligi_literojn (kuranta->sekv, kuranta1);
	kuranta1 = kuranta1->sekv;
	kuranta2 = kuranta2->sekv;
	lasta1 = kuranta1;
	if (ESTAS_D_LIT_PARENT (kuranta1->sekv)) {
		if (ESTAS_D_LIT_PARENT (kuranta2->sekv)) {
			kuranta->dat.cif = (kuranta1->dat.cif) /
				(kuranta2->dat.cif);
			kuranta1->dat.cif %= (kuranta2->dat.cif);
		};
	}
	else {
		kuranta1 = kuranta1->sekv;
		while (ESTAS_CIFER (kuranta1->sekv) && ESTAS_CIFER (kuranta2->sekv)) {
			kuranta1 = kuranta1->sekv;
			kuranta2 = kuranta2->sekv;
		};
		if (ESTAS_CIFER (kuranta2->sekv)) i = 0;
		if (ESTAS_CIFER ((kuranta2->sekv)->sekv)) i = -1;
		while (ESTAS_CIFER (kuranta1->sekv)) {
			kuranta1 = kuranta1->sekv;
			lasta1 = lasta1->sekv;
		};
		while (ESTAS_CIFER (lasta1->ant)) {
			while (kuranta1->dat.cif != 0) {
				k = ((unsigned long) kuranta1->dat.cif * 65536) +
					(unsigned long) (kuranta1->ant)->dat.cif;
				k /= (unsigned long) kuranta2->dat.cif + 1;
				arit_sumigi (kuranta, k);
				kuranta3 = kuranta1->ant;
				do {
					arit_subtrahi (kuranta3,
						(k % 65536) * (unsigned long) kuranta2->dat.cif);
					if (k >= 65536) {
						arit_subtrahi (kuranta3->sekv,
							(k / 65536) *
							(unsigned long) kuranta2->dat.cif);
					};
					kuranta3 = kuranta3->ant;
					kuranta2 = kuranta2->ant;
				} while (ESTAS_CIFER (kuranta2));
				kuranta2 = (KKP->ant)->ant;
			};
			kuranta1 = kuranta1->ant;
			lasta1 = lasta1->ant;
			krei_literon (kuranta->ant);
			(kuranta->ant)->sekv = kuranta;
			kuranta = kuranta->ant;
			kuranta->tip = CIFER;
			kuranta->dat.cif = 0;
			ligi_literojn (KKK->sekv, kuranta);
		};
		switch (i) {
			case 1:
				k = ((unsigned long) kuranta1->dat.cif * 65536) +
					(unsigned long) (kuranta1->ant)->dat.cif;
				while (k > (unsigned long) kuranta2->dat.cif) {
					k /= (unsigned long) kuranta2->dat.cif + 1;
					arit_sumigi (kuranta, k);
					kuranta3 = kuranta1->ant;
					do {
						arit_subtrahi (kuranta3,
							(k % 65536) *
							(unsigned long) kuranta2->dat.cif);
						if (k >= 65536) {
							arit_subtrahi (kuranta3->sekv,
								(k / 65536) *
								(unsigned long) kuranta2->dat.cif);
						};
						kuranta3 = kuranta3->ant;
						kuranta2 = kuranta2->ant;
					} while (ESTAS_CIFER (kuranta2));
					kuranta2 = (KKP->ant)->ant;
					k = ((unsigned long) kuranta1->dat.cif * 65536) +
						(unsigned long) (kuranta1->ant)->dat.cif;
				};
				kuranta1 = kuranta1->ant;
				if (kuranta1->dat.cif == kuranta2->dat.cif) {
					kuranta3 = kuranta1;
					while ((kuranta3->dat.cif == kuranta2->dat.cif) &&
					ESTAS_CIFER (kuranta3->ant)) {
						kuranta3 = kuranta3->ant;
						kuranta2 = kuranta2->ant;
					};
					if (kuranta3->dat.cif >= kuranta2->dat.cif) {
						arit_sumigi (kuranta, 1);
						while (kuranta1 != kuranta3) {
							kuranta1->dat.cif = 0;
							kuranta1 = kuranta1->ant;
						};
						while (!ESTAS_MD_LIT_PARENT (kuranta2)) {
							arit_subtrahi (kuranta1,
								(unsigned long) kuranta2->dat.cif);
							kuranta1 = kuranta1->ant;
							kuranta2 = kuranta2->ant;
						};
					};
				};
				break;
			
			case 0:
				kuranta2 = kuranta2->sekv;
				while (kuranta1->dat.cif > kuranta2->dat.cif) {
					k = (unsigned long) (kuranta1->dat.cif /
						(kuranta2->dat.cif + 1));
					arit_sumigi (kuranta, k);
					kuranta3 = kuranta1;
					do {
						arit_subtrahi (kuranta3,
							(k % 65536) *
							(unsigned long) kuranta2->dat.cif);
						if (k >= 65536) {
							arit_subtrahi (kuranta3->sekv,
								(k / 65536) *
								(unsigned long) kuranta2->dat.cif);
						};
						kuranta3 = kuranta3->ant;
						kuranta2 = kuranta2->ant;
					} while (ESTAS_CIFER (kuranta2));
					kuranta2 = (KKP->ant)->ant;
				};
				if (kuranta1->dat.cif == kuranta2->dat.cif) {
					kuranta3 = kuranta1;
					while ((kuranta3->dat.cif == kuranta2->dat.cif) &&
					ESTAS_CIFER (kuranta3->ant)) {
						kuranta3 = kuranta3->ant;
						kuranta2 = kuranta2->ant;
					};
					if (kuranta3->dat.cif >= kuranta2->dat.cif) {
						arit_sumigi (kuranta, 1);
						while (kuranta1 != kuranta3) {
							kuranta1->dat.cif = 0;
							kuranta1 = kuranta1->ant;
						};
						while (!ESTAS_MD_LIT_PARENT (kuranta2)) {
							arit_subtrahi (kuranta1,
								(unsigned long) kuranta2->dat.cif);
							kuranta1 = kuranta1->ant;
							kuranta2 = kuranta2->ant;
						};
					};
				};
				break;
			
			default:
				break;
		};
	};
	kuranta = (KKP->ant)->ant;
	do {
		detrui_literon (kuranta->sekv);
		kuranta = kuranta->ant;
	} while (!ESTAS_MD_LIT_PARENT (kuranta));
	ligi_literojn (kuranta->ant, KKP);
	detrui_literon (kuranta->sekv);
	detrui_literon (kuranta);
	kuranta = ((KKK->sekv)->dat.dual)->ant;
	while ((kuranta->dat.cif == 0) && ESTAS_CIFER (kuranta->ant)) {
		ligi_literojn (kuranta->ant, kuranta->sekv);
		detrui_literon (kuranta);
		kuranta = kuranta->ant;
	};
	kuranta = (KKP->ant)->ant;
	while ((kuranta->dat.cif == 0) && ESTAS_CIFER (kuranta->ant)) {
		ligi_literojn (kuranta->ant, kuranta->sekv);
		detrui_literon (kuranta);
		kuranta = kuranta->ant;
	};
	return;
}


/*
*	Функция "arit_div" производит целочисленное деление пары чисел,
*	заключённых между KKK и KKP.
*
*/

void arit_div (void) {
	LITERO *kuranta;
	
	arit_divmod ();
	kuranta = (KKP->ant)->ant;
	do {
		detrui_literon (kuranta->sekv);
		kuranta = kuranta->ant;
	} while (!ESTAS_MD_LIT_PARENT (kuranta));
	ligi_literojn (kuranta->ant, KKP);
	detrui_literon (kuranta->sekv);
	detrui_literon (kuranta);
	return;
}


/*
*	Функция "arit_mod" производит взятие остатка от целочисленного деления
*	пары чисел, заключённых между KKK и KKP.
*
*/

void arit_mod (void) {
	LITERO *kuranta;
	
	arit_divmod ();
	kuranta = (KKK->sekv)->sekv;
	do {
		detrui_literon (kuranta->ant);
		kuranta = kuranta->sekv;
	} while (!ESTAS_D_LIT_PARENT (kuranta));
	ligi_literojn (KKK, kuranta->sekv);
	detrui_literon (kuranta->ant);
	detrui_literon (kuranta);
	return;
}


/* ========================================================================== */


void arit_sumigi (LITERO *kuranta, unsigned long l) {
	unsigned long k;
	LITERO *kuranta1;
	
	kuranta1 = kuranta;
	k = l;
	do {
		k += (unsigned long) kuranta1->dat.cif;
		kuranta1->dat.cif = (unsigned int) (k % 65536);
		kuranta1 = kuranta1->sekv;
		k /= 65536;
	} while ((k != 0) && !ESTAS_D_LIT_PARENT (kuranta1));
	if (k != 0) {
		krei_literon ((kuranta1->ant)->sekv);
		((kuranta1->ant)->sekv)->ant = kuranta1->ant;
		ligi_literojn ((kuranta1->ant)->sekv, kuranta1);
		kuranta1 = kuranta1->ant;
		kuranta1->tip = CIFER;
		kuranta1->dat.cif = (unsigned int) k;
	};
	return;
}


void arit_subtrahi (LITERO *kuranta, unsigned long l) {
	unsigned long k, k1;
	LITERO *kuranta1;
	
	kuranta1 = kuranta;
	k = l;
	do {
		k1 = 65536 + (unsigned long) kuranta1->dat.cif - (k % 65536);
		kuranta1->dat.cif = (unsigned int) (k1 % 65536);
		kuranta1 = kuranta1->sekv;
		k /= 65536;
		if (k1 < 65536) k++;
	} while (k != 0);
	return;
}
