/*
*	bazo.c - общие функции библиотек.
*	© 2005, А.А.Владимиров.
*
*	Данная программа является свободным программным обеспечением.
*	Вы вправе распространять ее и/или модифицировать в соответствии
*	с условиями версии 2, либо, по вашему выбору, с условиями
*	более поздней версии Стандартной Общественной Лицензии GNU,
*	опубликованной Free Software Foundation.
*
*	Мы распространяем данную программу в надежде на то, что она будет вам
*	полезной, однако НЕ ПРЕДОСТАВЛЯЕМ НА НЕЕ НИКАКИХ ГАРАНТИЙ, в том числе
*	ГАРАНТИИ ТОВАРНОГО СОСТОЯНИЯ ПРИ ПРОДАЖЕ и ПРИГОДНОСТИ ДЛЯ ИСПОЛЬЗОВАНИЯ
*	В КОНКРЕТНЫХ ЦЕЛЯХ. Для получения более подробной информации ознакомьтесь
*	со Стандартной Общественной Лицензией GNU.
*
*	Вместе с данной программой вы должны были получить экземпляр
*	Стандартной Общественной Лицензии GNU. Если вы его не получили,
*	сообщите об этом в
*	
*	Free Software Foundation, Inc.,
*	59 Temple Place, Suite 330,
*	Boston, MA 02111-1307
*	USA
*
*/

# include "refal.h"


/* ========================================================================== */


extern void ref_eraro (int);


/* ========================================================================== */


/*
*	Функция "baz_krei_linion" записывает по адресу "linio" строку
*	(в кодировке UTF-8), заключённую между KKK и либо ближайшей к KKK справа
*	правой структурной скобкой, либо KKP. При этом переносимая в "linio"
*	часть поля зрения стирается.
*
*/

void baz_krei_linion (char *linio) {
	int i = 0, k = 0;
	unsigned long c = 0;
	LITERO *kuranta;
	
	kuranta = KKK->sekv;
	while ((kuranta != KKP) && (!ESTAS_D_STR_PARENT (kuranta))) {
		switch (kuranta->tip) {
			case MD_LIT_PARENT:
				c = 0;
				i = 0;
				break;
			
			case D_LIT_PARENT:
				i = 0;
				if ((i == 0) && (c < 128)) {
					if (k == LIN_LIM - 1) ref_eraro (5);
					linio [k] = (char) c;
					c = 0;
					k ++;
					i = 1;
				};
				if ((i == 0) && (c > 127) && (c < 2048)) {
					if (k == LIN_LIM - 1) ref_eraro (5);
					linio [k] = (char) ((c >> 6) + 192);
					c = (c << 26);
					k ++;
					i = 2;
				};
				if ((i == 0) && (c > 2047) && (c < 65536)) {
					if (k == LIN_LIM - 1) ref_eraro (5);
					linio [k] = (char) ((c >> 12) + 224);
					c = (c << 20);
					k ++;
					i = 3;
				};
				if ((i == 0) && (c > 65535) && (c < 2097152)) {
					if (k == LIN_LIM - 1) ref_eraro (5);
					linio [k] = (char) ((c >> 18) + 240);
					c = (c << 14);
					k ++;
					i = 4;
				};
				if ((i == 0) && (c > 2097151) && (c < 67108864)) {
					if (k == LIN_LIM - 1) ref_eraro (5);
					linio [k] = (char) ((c >> 24) + 248);
					c = (c << 8);
					k ++;
					i = 5;
				};
				if ((i == 0) && (c > 67108863) && (c < 2147483648u)) {
					if (k == LIN_LIM - 1) ref_eraro (5);
					linio [k] = (char) ((c >> 30) + 252);
					c = (c << 2);
					k ++;
					i = 6;
				};
				while (i != 1) {
					if (k == LIN_LIM - 1) ref_eraro (5);
					linio [k] = (char) ((c >> 26) + 128);
					c = (c << 6);
					k ++;
					i --;
				};
				break;
			
			case CIFER:
				switch (i) {
					case 0:
						c += kuranta->dat.cif;
						i ++;
						break;
					
					case 1:
						c += (kuranta->dat.cif) * 65536;
						if (c > 2147483647) ref_eraro (7);
						i ++;
						break;
					
					default:
						ref_eraro (7);
						break;
				};
				break;
			
			default:
				ref_eraro (8);
				break;
		};
		kuranta = kuranta->sekv;
		detrui_literon (kuranta->ant);
		ligi_literojn (KKK, kuranta);
	};
	linio [k] = '\0';
	return;
}


/*
*	Функция "baz_krei_bajtojn" записывает по адресу "linio" строку из байтов,
*	номера которых заключены между KKK и либо ближайшей к KKK справа
*	правой структурной скобкой, либо KKP. При этом переносимая в "linio"
*	часть поля зрения стирается.
*
*/

void baz_krei_bajtojn (char *linio) {
	int i = 0, k = 0;
	unsigned int c = 0;
	LITERO *kuranta;
	
	kuranta = KKK->sekv;
	while ((kuranta != KKP) && (!ESTAS_D_STR_PARENT (kuranta))) {
		switch (kuranta->tip) {
			case MD_LIT_PARENT:
				i = 0;
				break;
			
			case D_LIT_PARENT:
				if (c < 256) {
					if (k == LIN_LIM - 1) ref_eraro (5);
					linio [k] = (char) c;
					c = 0;
					k ++;
				};
				break;
			
			case CIFER:
				if (i) {
					ref_eraro (2);
				}
				else {
					c = kuranta->dat.cif;
					i ++;
				};
				break;
			
			default:
				ref_eraro (8);
				break;
		};
		kuranta = kuranta->sekv;
		detrui_literon (kuranta->ant);
		ligi_literojn (KKK, kuranta);
	};
	linio [k] = '\0';
	return;
}


/*
*	Функция "baz_preni_literon" считывает из строки "linio" букву
*	(в кодировке UTF-8), начало которой находится в позиции "*pozicio".
*
*	Возвращаемым значением функции является число типа (unsigned long),
*	представляющее собой номер считанной буквы в таблице юникода.
*
*	По окончании работы функции значением параметра "*pozicio" является
*	номер позиции, непосредственно следующей за считанной буквой.
*
*/

unsigned long baz_preni_literon (char *linio, int *pozicio) {
	int k;
	unsigned long l;
	
	if ((linio [*pozicio]) >> 7 == 0) {
		k = *pozicio;
		(*pozicio) ++;
		return (unsigned long) (linio [k]);
	};
	if (((linio [*pozicio]) >> 6 == 2) || ((linio [*pozicio]) == 254) ||
	((linio [*pozicio]) == 255)) ref_eraro (7);
	k = 32;
	while (((linio [*pozicio]) & k) != 0) k /= 2;
	switch (k) {
		case 32:
			k = 2;
			break;
		
		case 16:
			k = 3;
			break;
		
		case 8:
			k = 4;
			break;
		
		case 4:
			k = 5;
			break;
		
		case 2:
			k = 6;
			break;
	};
	l = (unsigned long) ((linio [*pozicio]) & (255 >> k));
	k = (*pozicio) + k;
	(*pozicio) ++;
	while (k != (*pozicio)) {
		if ((linio [*pozicio]) >> 6 != 2) ref_eraro (7);
		l *= 64;
		l += (unsigned long) ((linio [*pozicio]) & 63);
	 	(*pozicio) ++;
	};
	return l;
}


/*
*	Функция "baz_preni_bajton" считывает из строки "linio" байт
*	из позиции "*pozicio".
*
*	Возвращаемым значением функции является число типа (unsigned long),
*	представляющее собой номер считанного байта.
*
*	По окончании работы функции значение параметра "*pozicio"
*	инкрементируется.
*
*/

unsigned long baz_preni_bajton (char *linio, int *pozicio) {
	int k;
	
	k = *pozicio;
	(*pozicio) ++;
	return (unsigned long) (linio [k]);
}

