/*
*	dosieroj.c - библиотека средств ввода/вывода.
*	© 2005, А.А.Владимиров.
*
*	Данная программа является свободным программным обеспечением.
*	Вы вправе распространять ее и/или модифицировать в соответствии
*	с условиями версии 2, либо, по вашему выбору, с условиями
*	более поздней версии Стандартной Общественной Лицензии GNU,
*	опубликованной Free Software Foundation.
*
*	Мы распространяем данную программу в надежде на то, что она будет вам
*	полезной, однако НЕ ПРЕДОСТАВЛЯЕМ НА НЕЕ НИКАКИХ ГАРАНТИЙ, в том числе
*	ГАРАНТИИ ТОВАРНОГО СОСТОЯНИЯ ПРИ ПРОДАЖЕ и ПРИГОДНОСТИ ДЛЯ ИСПОЛЬЗОВАНИЯ
*	В КОНКРЕТНЫХ ЦЕЛЯХ. Для получения более подробной информации ознакомьтесь
*	со Стандартной Общественной Лицензией GNU.
*
*	Вместе с данной программой вы должны были получить экземпляр
*	Стандартной Общественной Лицензии GNU. Если вы его не получили,
*	сообщите об этом в
*	
*	Free Software Foundation, Inc.,
*	59 Temple Place, Suite 330,
*	Boston, MA 02111-1307
*	USA
*
*/

# include "refal.h"
# include <string.h>
# include <unistd.h>
# include <errno.h>


/*
*	Структура DOSIERO предназначена для организации списка открытых файлов.
*
*/

typedef struct dosiero {
	FILE *dat;			/* Указатель на открытый файл */
	int id;				/* Дескриптор файла */
	struct dosiero *ant;	/* Указатель на предыдущий элемент списка */
	struct dosiero *sekv;	/* Указатель на следующий элемент списка */
} DOSIERO;

DOSIERO* LDos;				/* Начало списка открытых файлов */
DOSIERO* LDosN;			/* Начало списка свободных элементов */

# define krei_dosieron(x) {\
	x = LDosN;\
	if (LDosN->sekv == NULL) {\
		LDosN = (DOSIERO*) malloc (sizeof (DOSIERO));\
		if (LDosN == NULL) ref_eraro (1);\
		LDosN->sekv = NULL;\
	}\
	else {\
		LDosN = LDosN->sekv;\
	};\
}

# define ligi_dosierojn(x,y) {x->sekv = y; y->ant = x;}
# define detrui_dosieron(x) {ligi_dosierojn (x, LDosN); LDosN = x;}


void dos_sxargi (void) __attribute__ ((constructor));

void dos_puts (void);
void dos_gets (void);
void dos_fopen (void);
void dos_fclose (void);
void dos_feof (void);
void dos_fputs (void);
void dos_fgets (void);


DOSIERO *dos_preni_dosieron (int);


/* ========================================================================== */


extern void ref_eraro (int);
extern void interp_aldoni_enkonstruitan (char* nomo, void (*funk) (void));
extern void baz_krei_linion (char*);
extern void baz_krei_bajtojn (char*);
extern unsigned long baz_preni_literon (char*, int*);
extern unsigned long baz_preni_bajton (char*, int*);


/* ========================================================================== */

void dos_sxargi () {
	
	LDosN = (DOSIERO*) malloc (sizeof (DOSIERO));
	if (LDosN == NULL) ref_eraro (1);
	LDosN->sekv = NULL;
	krei_dosieron (LDos);
	LDos->sekv = NULL;
	LDos->id = -1;
	interp_aldoni_enkonstruitan ("puts", dos_puts);
	interp_aldoni_enkonstruitan ("gets", dos_gets);
	interp_aldoni_enkonstruitan ("fopen", dos_fopen);
	interp_aldoni_enkonstruitan ("fclose", dos_fclose);
	interp_aldoni_enkonstruitan ("feof", dos_feof);
	interp_aldoni_enkonstruitan ("fputs", dos_fputs);
	interp_aldoni_enkonstruitan ("fgets", dos_fgets);
	return;
}

/* ========================================================================== */


/*
*	Функция "dos_puts" выводит строку, заключённую между KKK и KKP,
*	в канал "stderr".
*
*	Возможно использование перед выводимой строкой заключённого в скобки
*	параметра (типа кодировки).
*
*/

void dos_puts () {
	int i = 1;
	char *linio, *sign;
	LITERO *kuranta;
	
	linio = (char *) malloc (sizeof (char) * LIN_LIM);
	if (linio == NULL) ref_eraro (1);
	sign = (char *) malloc (sizeof (char) * LIN_LIM);
	if (sign == NULL) ref_eraro (1);
	if (ESTAS_MD_STR_PARENT (KKK->sekv)) {
		kuranta = (KKK->sekv)->sekv;
		detrui_literon (KKK->sekv);
		ligi_literojn (KKK, kuranta);
		baz_krei_linion (sign);
		kuranta = (KKK->sekv)->sekv;
		detrui_literon (KKK->sekv);
		ligi_literojn (KKK, kuranta);
		if (strcmp (sign, "utf-8") == 0) {
			i = 1;
		}
		else {
			if (strcmp (sign, "byte") == 0) {
				i = 0;
			}
			else {
				ref_eraro (8);
			};
		};
	};
	if (i) {
		baz_krei_linion (linio);
	}
	else {
		baz_krei_bajtojn (linio);
	};
	fputs (linio, stderr);
	free (linio);
	free (sign);
	return;
}


/*
*	Функция "dos_gets" считывает строку из канала "stdin".
*
*	Возможно использование заключённого в скобки параметра (типа кодировки).
*
*/

void dos_gets () {
	int pozicio, i = 1;
	unsigned long l;
	char *linio, *sign;
	LITERO *kuranta, *sekva;
	
	linio = (char *) malloc (sizeof (char) * LIN_LIM);
	if (linio == NULL) ref_eraro (1);
	sign = (char *) malloc (sizeof (char) * LIN_LIM);
	if (sign == NULL) ref_eraro (1);
	if (ESTAS_MD_STR_PARENT (KKK->sekv)) {
		kuranta = (KKK->sekv)->sekv;
		detrui_literon (KKK->sekv);
		ligi_literojn (KKK, kuranta);
		baz_krei_linion (sign);
		kuranta = (KKK->sekv)->sekv;
		detrui_literon (KKK->sekv);
		ligi_literojn (KKK, kuranta);
		if (strcmp (sign, "utf-8") == 0) {
			i = 1;
		}
		else {
			if (strcmp (sign, "byte") == 0) {
				i = 0;
			}
			else {
				ref_eraro (8);
			};
		};
	};
	if (KKK->sekv != KKP) ref_eraro (8);
	fgets (linio, LIN_LIM, stdin);
	pozicio = 0;
	while (linio [pozicio] != 10) {
		l = (i) ? baz_preni_literon (linio, &pozicio) :
		baz_preni_bajton (linio, &pozicio);
		krei_literon (kuranta);
		ligi_literojn (KKP->ant, kuranta);
		kuranta->tip = MD_LIT_PARENT;
		krei_literon (sekva);
		ligi_literojn (kuranta, sekva);
		sekva->tip = CIFER;
		sekva->dat.cif = (unsigned int) (l % 65536);
		krei_literon (kuranta);
		ligi_literojn (sekva, kuranta);
		kuranta->tip = D_LIT_PARENT;
		ligi_parentezojn (sekva->ant, kuranta);
		ligi_literojn (kuranta, KKP);
		l /= 65536;
		if (l != 0) {
			krei_literon (sekva);
			ligi_literojn (kuranta->ant, sekva);
			ligi_literojn (sekva, kuranta);
			sekva->tip = CIFER;
			sekva->dat.cif = (unsigned int) l;
		};
	};
	free (linio);
	free (sign);
	return;
};		


/*
*	Функция "dos_fopen" открывает файл, имя которого заключено между
*	"(KKK->sekv)->dat.dual" и KKP. Открытому файлу присваивается дескриптор,
*	указанный после "KKK->sekv". Между дескриптором и "(KKK->sekv)->dat.dual"
*	записывается тип открытия - одна из строк 'r', 'r+', 'w', 'w+', 'a', 'a+'.
*
*/

void dos_fopen () {
	int i, n;
	char *nomo, c [3];
	LITERO *kuranta;
	DOSIERO *dos;
	
	if (!ESTAS_MD_STR_PARENT (KKK->sekv)) ref_eraro (8);
	kuranta = (KKK->sekv)->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta)) ref_eraro (8);
	if (kuranta->dat.dual != (kuranta->sekv)->sekv) ref_eraro (8);
	kuranta = kuranta->sekv;
	n = (int) (kuranta->dat.cif);
	ligi_literojn (KKK->sekv, (kuranta->sekv)->sekv);
	detrui_literon (kuranta->sekv);
	detrui_literon (kuranta->ant);
	detrui_literon (kuranta);
	kuranta = (KKK->sekv)->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta)) ref_eraro (8);
	kuranta = kuranta->sekv;
	if ((kuranta->dat.cif != 97) && (kuranta->dat.cif != 114) &&
	(kuranta->dat.cif != 119)) ref_eraro (8);
	c [0] = (char) (kuranta->dat.cif);
	kuranta = kuranta->sekv;
	if (!ESTAS_D_LIT_PARENT (kuranta)) ref_eraro (8);
	ligi_literojn (KKK->sekv, kuranta->sekv);
	detrui_literon (kuranta->dat.dual);
	detrui_literon (kuranta->ant);
	detrui_literon (kuranta);
	kuranta = (KKK->sekv)->sekv;
	if (ESTAS_D_STR_PARENT (kuranta)) {
		c [1] = '\0';
		ligi_literojn (KKK, kuranta->sekv);
		detrui_literon (kuranta);
	}
	else {
		if (!ESTAS_MD_LIT_PARENT (kuranta)) ref_eraro (8);
		kuranta = kuranta->sekv;
		if (kuranta->dat.cif != 43) ref_eraro (8);
		c [1] = '+';
		c [2] = '\0';
		kuranta = kuranta->sekv;
		if (!ESTAS_D_LIT_PARENT (kuranta)) ref_eraro (8);
		if (!ESTAS_D_STR_PARENT (kuranta->sekv)) ref_eraro (8);
		ligi_literojn (KKK, (kuranta->sekv)->sekv);
		detrui_literon (kuranta->dat.dual);
		detrui_literon (kuranta->ant);
		detrui_literon (kuranta->sekv);
		detrui_literon (kuranta);
	};
	dos = LDos;
	i = 1;
	do {
		if (dos->sekv == NULL) {
			i = 0;
		}
		else {
			if ((dos->sekv)->id == n) i = 0;
			dos = dos->sekv;
		};
	} while (i);
	if (dos->id == n) {
		fclose (dos->dat);
	}
	else {
		krei_dosieron (dos->sekv);
		(dos->sekv)->ant = dos;
		dos = dos->sekv;
		dos->id = n;
		dos->sekv = NULL;
	};
	nomo = (char *) malloc (sizeof (char) * LIN_LIM);
	if (nomo == NULL) ref_eraro (1);
	baz_krei_linion (nomo);
	dos->dat = fopen (nomo, c);
	free (nomo);
	return;
}


/*
*	Функция "dos_fclose" закрывает файл, дескриптор которого указан между
*	KKK и KKP.
*
*/

void dos_fclose () {
	int n;
	LITERO *kuranta;
	DOSIERO *dos;
	
	if (!ESTAS_MD_LIT_PARENT (KKK->sekv)) ref_eraro (8);
	if ((KKK->sekv)->dat.dual != ((KKK->sekv)->sekv)->sekv) ref_eraro (8);
	kuranta = (KKK->sekv)->sekv;
	n = (int) (kuranta->dat.cif);
	ligi_literojn (KKK, (kuranta->sekv)->sekv);
	detrui_literon (kuranta->sekv);
	detrui_literon (kuranta->ant);
	detrui_literon (kuranta);
	if (KKK->sekv != KKP) ref_eraro (8);
	dos = dos_preni_dosieron (n);
	if (dos != NULL) {
		fclose (dos->dat);
		if (dos->sekv == NULL) {
			(dos->ant)->sekv = NULL;
			detrui_dosieron (dos);
		}
		else {
			ligi_dosierojn (dos->ant, dos->sekv);
			detrui_dosieron (dos);
		};
	};
	return;
}


/*
*	Функция "dos_feof" проверяет возможность чтения из файла, дескриптор
*	которого заключён между KKK и KKP.
*
*/

void dos_feof () {
	int n, c;
	LITERO *kuranta;
	DOSIERO *dos;
	
	if (!ESTAS_MD_LIT_PARENT (KKK->sekv)) ref_eraro (8);
	if ((KKK->sekv)->dat.dual != ((KKK->sekv)->sekv)->sekv) ref_eraro (8);
	kuranta = (KKK->sekv)->sekv;
	if ((kuranta->sekv)->sekv != KKP) ref_eraro (8);
	n = (int) (kuranta->dat.cif);
	dos = dos_preni_dosieron (n);
	if (dos != NULL) {
		c = getc (dos->dat);
		if (c == EOF) {
			kuranta->dat.cif = 0;
		}
		else {
			kuranta->dat.cif = 1;
			ungetc (c, dos->dat);
		};
	}
	else {
		kuranta->dat.cif = 0;
	};
	return;
}


/*
*	Функция "dos_fputs" выводит строку, заключённую между
*	"(KKK->sekv)->dat.dual" и KKP, в файл, дескриптор которого указан между
*	"KKK->sekv" и "(KKK->sekv)->dat.dual".
*
*	Возможно использование после дескриптора файла дополнительного параметра
*	(типа кодировки).
*
*/

void dos_fputs () {
	int n, i = 1;
	char *linio, *sign;
	LITERO *kuranta;
	DOSIERO *dos;
	
	linio = (char *) malloc (sizeof (char) * LIN_LIM);
	if (linio == NULL) ref_eraro (1);
	sign = (char *) malloc (sizeof (char) * LIN_LIM);
	if (sign == NULL) ref_eraro (1);
	if (!ESTAS_MD_STR_PARENT (KKK->sekv)) ref_eraro (8);
	kuranta = (KKK->sekv)->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta)) ref_eraro (8);
	if (kuranta->dat.dual != (kuranta->sekv)->sekv) ref_eraro (8);
	kuranta = kuranta->sekv;
	n = (int) (kuranta->dat.cif);
	ligi_literojn (KKK, (kuranta->sekv)->sekv);
	detrui_literon (kuranta->sekv);
	detrui_literon ((kuranta->ant)->ant);
	detrui_literon (kuranta->ant);
	detrui_literon (kuranta);
	baz_krei_linion (sign);
	kuranta = (KKK->sekv)->sekv;
	detrui_literon (KKK->sekv);
	ligi_literojn (KKK, kuranta);
	if ((strcmp (sign, "utf-8") == 0) || (strcmp (sign, "") == 0)) {
		i = 1;
	}
	else {
		if (strcmp (sign, "byte") == 0) {
			i = 0;
		}
		else {
			ref_eraro (8);
		};
	};
	if (i) {
		baz_krei_linion (linio);
	}
	else {
		baz_krei_bajtojn (linio);
	};
	dos = dos_preni_dosieron (n);
	if (dos != NULL) fputs (linio, dos->dat);
	free (linio);
	free (sign);
	return;
}


/*
*	Функция "dos_fgets" считывает строку из файла, дескриптор которого указан
*	между "KKK->sekv" и "(KKK->sekv)->dat.dual".
*
*	Возможно использование после дескриптора файла дополнительного параметра
*	(типа кодировки).
*
*/

void dos_fgets () {
	int n, pozicio, i = 1;
	unsigned long l;
	char *linio, *sign;
	LITERO *kuranta, *sekva;
	DOSIERO *dos;
	
	linio = (char *) malloc (sizeof (char) * LIN_LIM);
	if (linio == NULL) ref_eraro (1);
	sign = (char *) malloc (sizeof (char) * LIN_LIM);
	if (sign == NULL) ref_eraro (1);
	if (!ESTAS_MD_STR_PARENT (KKK->sekv)) ref_eraro (8);
	kuranta = (KKK->sekv)->sekv;
	if (!ESTAS_MD_LIT_PARENT (kuranta)) ref_eraro (8);
	if (kuranta->dat.dual != (kuranta->sekv)->sekv) ref_eraro (8);
	kuranta = kuranta->sekv;
	n = (int) (kuranta->dat.cif);
	ligi_literojn (KKK, (kuranta->sekv)->sekv);
	detrui_literon (kuranta->sekv);
	detrui_literon ((kuranta->ant)->ant);
	detrui_literon (kuranta->ant);
	detrui_literon (kuranta);
	baz_krei_linion (sign);
	kuranta = (KKK->sekv)->sekv;
	detrui_literon (KKK->sekv);
	ligi_literojn (KKK, kuranta);
	if ((strcmp (sign, "utf-8") == 0) || (strcmp (sign, "") == 0)) {
		i = 1;
	}
	else {
		if (strcmp (sign, "byte") == 0) {
			i = 0;
		}
		else {
			ref_eraro (8);
		};
	};
	if (KKK->sekv != KKP) ref_eraro (8);
	dos = dos_preni_dosieron (n);
	if (dos != NULL) {
		fgets (linio, LIN_LIM, dos->dat);
	}
	else {
		linio [0] = 10;
	};
	pozicio = 0;
	while (linio [pozicio] != 10) {
		l = (i) ? baz_preni_literon (linio, &pozicio) :
		baz_preni_bajton (linio, &pozicio);
		krei_literon (kuranta);
		ligi_literojn (KKP->ant, kuranta);
		kuranta->tip = MD_LIT_PARENT;
		krei_literon (sekva);
		ligi_literojn (kuranta, sekva);
		sekva->tip = CIFER;
		sekva->dat.cif = (unsigned int) (l % 65536);
		krei_literon (kuranta);
		ligi_literojn (sekva, kuranta);
		kuranta->tip = D_LIT_PARENT;
		ligi_parentezojn (sekva->ant, kuranta);
		ligi_literojn (kuranta, KKP);
		l /= 65536;
		if (l != 0) {
			krei_literon (sekva);
			ligi_literojn (kuranta->ant, sekva);
			ligi_literojn (sekva, kuranta);
			sekva->tip = CIFER;
			sekva->dat.cif = (unsigned int) l;
		};
	};
	free (linio);
	free (sign);
	return;
};		


/* ========================================================================== */


DOSIERO *dos_preni_dosieron (int n) {
	int i;
	DOSIERO *dos;
	
	dos = LDos;
	i = 1;
	do {
		if (dos->sekv == NULL) {
			i = 0;
		}
		else {
			if ((dos->sekv)->id == n) i = 0;
			dos = dos->sekv;
		};
	} while (i);
	if (dos->id == n) return dos;
	else return NULL;
}

