/*
*	refal.h - описание типов данных представляемой реализации языка Рефал.
*	© 2005, А.А.Владимиров.
*
*	Данная программа является свободным программным обеспечением.
*	Вы вправе распространять ее и/или модифицировать в соответствии
*	с условиями версии 2, либо, по вашему выбору, с условиями
*	более поздней версии Стандартной Общественной Лицензии GNU,
*	опубликованной Free Software Foundation.
*
*	Мы распространяем данную программу в надежде на то, что она будет вам
*	полезной, однако НЕ ПРЕДОСТАВЛЯЕМ НА НЕЕ НИКАКИХ ГАРАНТИЙ, в том числе
*	ГАРАНТИИ ТОВАРНОГО СОСТОЯНИЯ ПРИ ПРОДАЖЕ и ПРИГОДНОСТИ ДЛЯ ИСПОЛЬЗОВАНИЯ
*	В КОНКРЕТНЫХ ЦЕЛЯХ. Для получения более подробной информации ознакомьтесь
*	со Стандартной Общественной Лицензией GNU.
*
*	Вместе с данной программой вы должны были получить экземпляр
*	Стандартной Общественной Лицензии GNU. Если вы его не получили,
*	сообщите об этом в
*	
*	Free Software Foundation, Inc.,
*	59 Temple Place, Suite 330,
*	Boston, MA 02111-1307
*	USA
*
*/

# include <stdio.h>
# include <stdlib.h>

/*
*	Структура LITERO предназначена для записи элементов поля зрения
*	рефал-системы. Она представляет собой основной в этой системе тип данных.
*	
*	Типы литер могут быть следующими:
*
*	0) Макроцифра. Значением является целое число "cif" типа
*	(unsigned int), величина которого заключена между 0 и 65535. 
*
*	1) Левая литероидная скобка. Значением является ссылка "dual" типа
*	(LITERO *) на соответствующую правую скобку.
*
*	2) Правая литероидная скобка. Значением является ссылка "dual" типа
*	(LITERO *) на соответствующую левую скобку.
*
*	# Литероидные скобки ограничивают часть выражения, образующую
*	# неделимый символ. Таким образом, число символов оказывается
*	# потенциально бесконечным. Этот способ введения символов используется
*	# в математической логике при кодировании слов в произвольном алфавите
*	# словами в заранее фиксированном двух- (или более) буквенном алфавите.
*	# В представленной реализации языка Рефал использован подход, аналогичный
*	# замене букв произвольного алфавита литероидами (a), (aa), (aaa) и т.д.,
*	# представляющими собой слова в трёхбуквенном алфавите a().
*
*	3) Левая структурная скобка. Значением является ссылка "dual" типа
*	(LITERO *) на соответствующую правую скобку.
*
*	4) Правая структурная скобка. Значением является ссылка "dual" типа
*	(LITERO *) на соответствующую левую скобку.
*
*	5) Конкретизационный символ "k". Значением является ссылка "funk" типа
*	(FUNKCIO *) на подлежащую применению функцию.
*	
*	6) Конкретизационная точка. Значением является ссылка "dual" типа
*	(LITERO *) на соответствующий конкретизационный символ "k".
*	
*	7) Переменная. Значением является ссылка "var" типа
*	(VARIABLO *) на описание этой переменной.
*
*/

typedef struct litero {
	int tip;				/* Номер типа символа */
	union {

		unsigned int cif;		/* Макроцифра */
		struct litero *dual;	/* Ссылка на дуальную скобку */
		struct variablo *var;	/* Ссылка на описание переменной */
		struct funkcio *funk;	/* Ссылка на функцию */

	} dat;
	struct litero *ant;		/* Ссылка на предыдущую литеру */
	struct litero *sekv;	/* Ссылка на следующую литеру */
} LITERO;


# define CIFER 0			/* Макроцифра */
# define MD_LIT_PARENT 1		/* Левая литероидная скобка */
# define D_LIT_PARENT 2		/* Правая литероидная скобка */
# define MD_STR_PARENT 3		/* Левая структурная скобка */
# define D_STR_PARENT 4		/* Правая структурная скобка */
# define KONK_K 5			/* Конкретизационный символ "k" */
# define KONK_P 6			/* Конкретизационная точка */
# define VAR 7				/* Переменная */

# define ESTAS_CIFER(x) (x->tip == 0)
# define ESTAS_MD_LIT_PARENT(x) (x->tip == 1)
# define ESTAS_D_LIT_PARENT(x) (x->tip == 2)
# define ESTAS_MD_STR_PARENT(x) (x->tip == 3)
# define ESTAS_D_STR_PARENT(x) (x->tip == 4)
# define ESTAS_KONK_K(x) (x->tip == 5)
# define ESTAS_KONK_P(x) (x->tip == 6)
# define ESTAS_VAR(x) (x->tip == 7)


/*
*	Структура VARIABLO предназначена для описания переменных
*	в формулах подстановки.
*	
*	Типы переменных могут быть следующими:
*	
*	0) 'Активная' s-переменная. Данный тип отвечает самому левому вхождению
*	переменной с данным именем (если речь идёт об образце), или переменной,
*	значение которой представляет собой копию некоторого участка - а не сам
*	этот участок! - исходного выражения (если речь идёт о правой части
*	формулы подстановки). Значение элемента "ref" имеет смысл (и используется)
*	только для переменной из правой части формулы подстановки: в этом случае
*	оно представляет собой ссылку на переменную из образца, значение которой
*	подлежит копированию.
*
*	1) 'Пассивная' s-переменная. Данный тип отвечает вхождениям переменных,
*	левее которых уже имелись вхождения переменной с тем же именем (если речь
*	идёт об образце), или переменной, значение которой представляет собой
*	некоторый участок - а не его копию! - исходного выражения (если речь идёт
*	о правой части формулы подстановки). Значение элемента "ref" представляет
*	собой ссылку на самое левое вхождение переменной с данным именем в образец
*	(для переменной из образца), или ссылку на переменную из образца, значение
*	которой подлежит переносу.
*	
*	2) 'Активная' w-переменная. Данный тип отвечает самому левому вхождению
*	переменной с данным именем (если речь идёт об образце), или переменной,
*	значение которой представляет собой копию некоторого участка - а не сам
*	этот участок! - исходного выражения (если речь идёт о правой части
*	формулы подстановки). Значение элемента "ref" имеет смысл (и используется)
*	только для переменной из правой части формулы подстановки: в этом случае
*	оно представляет собой ссылку на переменную из образца, значение которой
*	подлежит копированию.
*
*	3) 'Пассивная' w-переменная. Данный тип отвечает вхождениям переменных,
*	левее которых уже имелись вхождения переменной с тем же именем (если речь
*	идёт об образце), или переменной, значение которой представляет собой
*	некоторый участок - а не его копию! - исходного выражения (если речь идёт
*	о правой части формулы подстановки). Значение элемента "ref" представляет
*	собой ссылку на самое левое вхождение переменной с данным именем в образец
*	(для переменной из образца), или ссылку на переменную из образца, значение
*	которой подлежит переносу.
*	
*	4) 'Активная' e-переменная. Данный тип отвечает самому левому вхождению
*	переменной с данным именем - при условии, что значение этой переменной
*	не определяется положением её левой границы - (если речь идёт об образце),
*	или переменной, значение которой представляет собой копию некоторого
*	участка - а не сам этот участок! - исходного выражения (если речь идёт
*	о правой части формулы подстановки). Значение элемента "ref" имеет смысл
*	(и используется) только для переменной из правой части формулы подстановки:
*	в этом случае оно представляет собой ссылку на переменную из образца,
*	значение которой подлежит копированию.
*	
*	5) 'Полуактивная' e-переменная. Данный тип используется только в образцах
*	и отвечает самому левому вхождению переменной с данным именем при условии,
*	что значение этой переменной однозначно определяется положением её левой
*	границы. Значение элемента "ref" игнорируется.
*	
*	6) 'Пассивная' e-переменная. Данный тип отвечает вхождениям переменных,
*	левее которых уже имелись вхождения переменной с тем же именем (если речь
*	идёт об образце), или переменной, значение которой представляет собой
*	некоторый участок - а не его копию! - исходного выражения (если речь идёт
*	о правой части формулы подстановки). Значение элемента "ref" представляет
*	собой ссылку на самое левое вхождение переменной с данным именем в образец
*	(для переменной из образца), или ссылку на переменную из образца, значение
*	которой подлежит переносу.
	
*	Значением элемента "ant" является ссылка на литеру из поля зрения,
*	непосредственно предшествую значению переменной. Значением элемента "sekv"
*	является ссылка на литеру из поля зрения, непосредственно следующую
*	за значением переменной. Указанные значения устанавливаются
*	в ходе сопоставления.
*	
*	# ПРИМЕР: Рассмотрим формулу подстановки
*	#
*	#		s/1/ (s/1/) e/2/ e/3/ = s/1/ e/2/ e/2/
*	#		1     2     3    4      5    6    7
*	#
*	# Переменные 1 и 3 являются 'активными' переменными образца;
*	# переменная 4 является 'полуактивной'; переменная 2 'пассивна' и отвечает
*	# 'активной' переменной 1. Переменные 5 и 6 являются 'пассивными'
*	# переменными из правой части - значение переменной 5 может быть получено
*	# переносом значения переменной 1 (или 2), а значение переменной 6
*	# получается переносом значения переменной 3. Напротив, переменная 7
*	# является 'активной' (ибо e/2/ входит в образец лишь единожды) -
*	# и для построения её значения необходимо произвести копирование
*	# значения переменной 3.
*
*/

typedef struct variablo {
	int tip;				/* Номер типа переменной */
	struct variablo *ref;	/* Ссылка на другую переменную */
	struct litero *ant;		/* Ссылка на предшествующую значению литеру */
	struct litero *sekv;	/* Ссылка на следующую за значением литеру */
} VARIABLO;


# define S_AGEM 0			/* 'Активная' s-переменная */
# define S_MALAGEM 1		/* 'Пассивная' s-переменная */
# define W_AGEM 2			/* 'Активная' w-переменная */
# define W_MALAGEM 3		/* 'Пассивная' w-переменная */
# define E_AGEM 4			/* 'Активная' e-переменная */
# define E_DUONAGEM 5		/* 'Полуактивная' e-переменная */
# define E_MALAGEM 6		/* 'Пассивная' e-переменная */

# define ESTAS_S_AGEM(x) (x->tip == 0)
# define ESTAS_S_MALAGEM(x) (x->tip == 1)
# define ESTAS_W_AGEM(x) (x->tip == 2)
# define ESTAS_W_MALAGEM(x) (x->tip == 3)
# define ESTAS_E_AGEM(x) (x->tip == 4)
# define ESTAS_E_DUONAGEM(x) (x->tip == 5)
# define ESTAS_E_MALAGEM(x) (x->tip == 6)


/*
*	Структура SUBSTITUO предназначена для записи формул подстановки.
*
*	Значением элемента "mod" является ссылка на первую литеру образца.
*	
*	# Первой литерой образца должен быть конкретизационный символ "k",
*	# а последней - конкретизационная точка (между которыми заключается
*	# 'действительный' образец). При этом, вопреки общему правилу, данные
*	# используемого конкретизационного символа "k" должны иметь тип "dual"
*	# (вместо "funk"): первому и последнему символам образца полагается
*	# указывать друг на друга.
*	
*	Значением элемента "rez" является ссылка на первую литеру правой части
*	формулы подстановки, или "NULL" (если правая часть формулы подстановки
*	пуста).
*
*	# Последняя литера непустой правой части должна подчиняться условию
*	# "x->sekv == NULL".
*	
*	Значением элемента "sekv" является ссылка на следующую формулу подстановки,
*	или "NULL" (если рассматриваемая формула является последней в списке).
*
*/


typedef struct substituo {
	struct litero *mod;		/* Ссылка на образец (левую часть) */
	struct litero *rez;		/* Ссылка на правую часть */
	struct substituo *sekv;	/* Ссылка на следующую формулу подстановки */
} SUBSTITUO;


/*
*	Структура FUNKCIO предназначена для записи функций рефал-машины.
*
*	Функция может быть встроенной или 'чисто рефальской'. Если функция
*	является встроенной, значение элемента "enk" равно 1; иначе оно равно 0.
*
*	Встроенные функции должны быть написаны на языке C в следующем формате:
*
*		void ИМЯ_ФУНКЦИИ (void);
*
*	Значением элемента "f" является ссылка на функцию указанного формата.
*
*	'Чисто рефальская' функция определяется списком формул подстановки.
*	Значением элемента "s" является ссылка на первую формулу подстановки из
*	списка, отвечающего рассматриваемой функции.
*
*	Каждая функция (исключая единственную внутреннюю функцию, останавливающую
*	рефал-машину) имеет имя. Предполагается, что таковым является некоторое
*	слово юникода; оно сохраняется в виде массива чисел типа (unsigned long).
*
*	Функции рефал-машины образуют список. Для его организации предназначен
*	элемент "sekv", значением которого является ссылка на функцию,
*	следующую за рассматриваемой, либо "NULL", если рассматриваемая функция
*	является последней в списке.
*	
*/


typedef struct funkcio {
	int enk;					/* Тип функции */
	unsigned long *nom;			/* Имя функции */
	union {
	
		void (*f) (void);		/* Ссылка на встроенную функцию */
		struct substituo *s;	/* Ссылка на первую формулу подстановки */

	} dat;
	struct funkcio *sekv;		/* Ссылка на следующую функцию */
} FUNKCIO;


extern int ARGC;			/* Первый аргумент C-функции "main" */
extern char **ARGV;			/* Второй аргумент C-функции "main" */

extern int NOM_LIM;			/* Максимальная длина имени */
extern int LIN_LIM;			/* Максимальная длина строки */
extern int VAR_LIM;			/* Максимальное число переменных
						в формуле подстановки */

extern LITERO *LLM;			/* Начало списка свободных литер */
extern FUNKCIO *LF;			/* Начало списка функций */
extern LITERO *LKP;			/* Начало списка конкретизаций */
extern LITERO *KKK;			/* Текущий конкретизационный символ "k" */
extern LITERO *KKP;			/* Текущая конкретизационная точка */
extern SUBSTITUO *KFS;		/* Текущая формула подстановки */


# define ligi_literojn(x,y) {x->sekv = y; y->ant = x;}
# define ligi_parentezojn(x,y) {x->dat.dual = y; y->dat.dual = x;}


# define krei_literon(x) {\
	x = LLM;\
	if (LLM->sekv == NULL) {\
		LLM = (LITERO *) malloc (sizeof (LITERO));\
		if (LLM == NULL) ref_eraro (1);\
		LLM->sekv = NULL;\
	}\
	else {\
		LLM = LLM->sekv;\
	};\
}
		

# define detrui_literon(x) {ligi_literojn (x, LLM); LLM = x;}

