/*
*	sistemo.c - библиотека системных функций.
*	© 2005, А.А.Владимиров.
*
*	Данная программа является свободным программным обеспечением.
*	Вы вправе распространять ее и/или модифицировать в соответствии
*	с условиями версии 2, либо, по вашему выбору, с условиями
*	более поздней версии Стандартной Общественной Лицензии GNU,
*	опубликованной Free Software Foundation.
*
*	Мы распространяем данную программу в надежде на то, что она будет вам
*	полезной, однако НЕ ПРЕДОСТАВЛЯЕМ НА НЕЕ НИКАКИХ ГАРАНТИЙ, в том числе
*	ГАРАНТИИ ТОВАРНОГО СОСТОЯНИЯ ПРИ ПРОДАЖЕ и ПРИГОДНОСТИ ДЛЯ ИСПОЛЬЗОВАНИЯ
*	В КОНКРЕТНЫХ ЦЕЛЯХ. Для получения более подробной информации ознакомьтесь
*	со Стандартной Общественной Лицензией GNU.
*
*	Вместе с данной программой вы должны были получить экземпляр
*	Стандартной Общественной Лицензии GNU. Если вы его не получили,
*	сообщите об этом в
*	
*	Free Software Foundation, Inc.,
*	59 Temple Place, Suite 330,
*	Boston, MA 02111-1307
*	USA
*
*/

# include "refal.h"
# include <sys/types.h>
# include <unistd.h>

void sis_sxargi (void) __attribute__ ((constructor));

void sis_system (void);
void sis_exit (void);
void sis_getpid (void);
void sis_getppid (void);
void sis_chdir (void);
void sis_getcwd (void);


/* ========================================================================== */


extern void ref_eraro (int);
extern void interp_aldoni_enkonstruitan (char* nomo, void (*funk) (void));
extern void baz_krei_linion (char*);
extern unsigned long baz_preni_literon (char*, int*);


/* ========================================================================== */

void sis_sxargi () {
	
	interp_aldoni_enkonstruitan ("system", sis_system);
	interp_aldoni_enkonstruitan ("exit", sis_exit);
	interp_aldoni_enkonstruitan ("getpid", sis_getpid);
	interp_aldoni_enkonstruitan ("getppid", sis_getppid);
	interp_aldoni_enkonstruitan ("chdir", sis_chdir);
	interp_aldoni_enkonstruitan ("getcwd", sis_getcwd);
	return;
}

/* ========================================================================== */


/*
*	Функция "sis_sistem" исполняет команду оболочки, записанную между
*	KKK и KKP.
*
*	Результатом работы функции является пара чисел, из которых первое
*	равно 0, если вызванная C-функция system возвратила значение -1, и
*	равно 1 иначе, а второе представляет собой возвращённое значение
*	C-функции system (в случае неотрицательности последнего).
*
*/

void sis_system () {
	int n;
	char *linio;
	LITERO *kuranta, *sekva, *antauxa;
	
	linio = (char *) malloc (sizeof (char) * LIN_LIM);
	if (linio == NULL) ref_eraro (1);
	baz_krei_linion (linio);
	n = system (linio);
	free (linio);
	krei_literon (antauxa);
	krei_literon (kuranta);
	krei_literon (sekva);
	ligi_literojn (KKK, antauxa);
	ligi_literojn (antauxa, kuranta);
	ligi_literojn (kuranta, sekva);
	kuranta->tip = CIFER;
	kuranta->dat.cif = (n == -1) ? 0 : 1;
	antauxa->tip = MD_LIT_PARENT;
	sekva->tip = D_LIT_PARENT;
	ligi_parentezojn (antauxa, sekva);
	krei_literon (antauxa);
	ligi_literojn (sekva, antauxa);
	krei_literon (kuranta);
	krei_literon (sekva);
	ligi_literojn (antauxa, kuranta);
	ligi_literojn (kuranta, sekva);
	ligi_literojn (sekva, KKP);
	kuranta->tip = CIFER;
	kuranta->dat.cif = (n == -1) ? 0 : (unsigned int) n;
	antauxa->tip = MD_LIT_PARENT;
	sekva->tip = D_LIT_PARENT;
	ligi_parentezojn (antauxa, sekva);
	return;
}


/*
*	Функция "sis_exit" осуществляет выход из программы.
*
*/

void sis_exit () {
	int n;
	LITERO *kuranta;
	
	if (!ESTAS_MD_LIT_PARENT (KKK->sekv)) ref_eraro (8);
	if ((KKK->sekv)->dat.dual != ((KKK->sekv)->sekv)->sekv) ref_eraro (8);
	kuranta = (KKK->sekv)->sekv;
	n = (int) (kuranta->dat.cif);
	ligi_literojn (KKK, (kuranta->sekv)->sekv);
	detrui_literon (kuranta->sekv);
	detrui_literon (kuranta->ant);
	detrui_literon (kuranta);
	if (KKK->sekv != KKP) ref_eraro (8);
	exit (n);
	return;
}


/*
*	Функция "sis_getpid" определяет номер текущего процесса.
*
*/

void sis_getpid () {
	LITERO *kuranta;
	
	if (KKK->sekv != KKP) ref_eraro (8);
	krei_literon (KKK->sekv);
	(KKK->sekv)->tip = MD_LIT_PARENT;
	(KKK->sekv)->ant = KKK;
	krei_literon (KKP->ant);
	(KKP->ant)->tip = D_LIT_PARENT;
	(KKP->ant)->sekv = KKP;
	ligi_parentezojn (KKK->sekv, KKP->ant);
	krei_literon (kuranta);
	kuranta->tip = CIFER;
	kuranta->dat.cif = (unsigned int) getpid ();
	ligi_literojn (KKK->sekv, kuranta);
	ligi_literojn (kuranta, KKP->ant);
	return;
}


/*
*	Функция "sis_getppid" определяет номер родительского процесса.
*
*/

void sis_getppid () {
	LITERO *kuranta;
	
	if (KKK->sekv != KKP) ref_eraro (8);
	krei_literon (KKK->sekv);
	(KKK->sekv)->tip = MD_LIT_PARENT;
	(KKK->sekv)->ant = KKK;
	krei_literon (KKP->ant);
	(KKP->ant)->tip = D_LIT_PARENT;
	(KKP->ant)->sekv = KKP;
	ligi_parentezojn (KKK->sekv, KKP->ant);
	krei_literon (kuranta);
	kuranta->tip = CIFER;
	kuranta->dat.cif = (unsigned int) getppid ();
	ligi_literojn (KKK->sekv, kuranta);
	ligi_literojn (kuranta, KKP->ant);
	return;
}


/*
*	Функция "sis_chdir" изменяет рабочую директорию на ту, имя которой
*	заключено между KKK и KKP.
*
*/

void sis_chdir () {
	char *linio;
	
	linio = (char *) malloc (sizeof (char) * LIN_LIM);
	if (linio == NULL) ref_eraro (1);
	baz_krei_linion (linio);
	chdir (linio);
	free (linio);
	return;
}


/*
*	Функция "sis_getcwd" определяет имя рабочей директории.
*
*/

void sis_getcwd () {
	int pozicio;
	unsigned long l;
	char *linio;
	LITERO *kuranta, *sekva;
	
	if (KKK->sekv != KKP) ref_eraro (8);
	linio = (char *) malloc (sizeof (char) * LIN_LIM);
	if (linio == NULL) ref_eraro (1);
	getcwd (linio, LIN_LIM);
	pozicio = 0;
	while (linio [pozicio] != '\0') {
		l = baz_preni_literon (linio, &pozicio);
		krei_literon (kuranta);
		ligi_literojn (KKP->ant, kuranta);
		kuranta->tip = MD_LIT_PARENT;
		krei_literon (sekva);
		ligi_literojn (kuranta, sekva);
		sekva->tip = CIFER;
		sekva->dat.cif = (unsigned int) (l % 65536);
		krei_literon (kuranta);
		ligi_literojn (sekva, kuranta);
		kuranta->tip = D_LIT_PARENT;
		ligi_parentezojn (sekva->ant, kuranta);
		ligi_literojn (kuranta, KKP);
		l /= 65536;
		if (l != 0) {
			krei_literon (sekva);
			ligi_literojn (kuranta->ant, sekva);
			ligi_literojn (sekva, kuranta);
			sekva->tip = CIFER;
			sekva->dat.cif = (unsigned int) l;
		};
	};
	free (linio);
	return;
};		


