/*
*	refal.c - функция "main" представляемой реализации языка Рефал.
*	© 2005, А.А.Владимиров.
*
*	Данная программа является свободным программным обеспечением.
*	Вы вправе распространять ее и/или модифицировать в соответствии
*	с условиями версии 2, либо, по вашему выбору, с условиями
*	более поздней версии Стандартной Общественной Лицензии GNU,
*	опубликованной Free Software Foundation.
*
*	Мы распространяем данную программу в надежде на то, что она будет вам
*	полезной, однако НЕ ПРЕДОСТАВЛЯЕМ НА НЕЕ НИКАКИХ ГАРАНТИЙ, в том числе
*	ГАРАНТИИ ТОВАРНОГО СОСТОЯНИЯ ПРИ ПРОДАЖЕ и ПРИГОДНОСТИ ДЛЯ ИСПОЛЬЗОВАНИЯ
*	В КОНКРЕТНЫХ ЦЕЛЯХ. Для получения более подробной информации ознакомьтесь
*	со Стандартной Общественной Лицензией GNU.
*
*	Вместе с данной программой вы должны были получить экземпляр
*	Стандартной Общественной Лицензии GNU. Если вы его не получили,
*	сообщите об этом в
*	
*	Free Software Foundation, Inc.,
*	59 Temple Place, Suite 330,
*	Boston, MA 02111-1307
*	USA
*
*/

# include "refal.h"
# include <libintl.h>


int NOM_LIM = 32;	/* Максимальная длина имени */
int LIN_LIM = 2048;	/* Максимальная длина строки */
int VAR_LIM = 64;	/* Максимальное число переменных в формуле подстановки */


int ARGC;			/* Первый аргумент C-функции "main" */
char **ARGV;		/* Второй аргумент C-функции "main" */


LITERO *LLM;		/* Начало списка свободных литер */
FUNKCIO *LF;		/* Начало списка функций */
LITERO *LKP;		/* Начало списка конкретизаций */
LITERO *KKK;		/* Текущий конкретизационный символ "k" */
LITERO *KKP;		/* Текущая конкретизационная точка */
SUBSTITUO *KFS;	/* Текущая формула подстановки */


/* ========================================================================== */


extern void ref_lancxi (void);
extern void ref_cxesigi (void);
extern void ref_eraro (int);
extern void interp_krei_masxinon (char*);


/* ========================================================================== */

/*
*	Функция "main" устанавливает начальное состояние рефал-машины:
*	1) Создаётся список функций, содержащий две функции:
*		a) безымянную функцию, останавливающую машину (ниже условно
*		   называемую "$halt");
*		b) функцию "$start", запускающую машину.
*	2) Создаётся поле зрения вида
*		k/$halt/ k/$start/ . .
*		1        2         3 4
*	3) Создаётся список конкретизаций, содержащий ссылки на литеры 3 и 4
*	(см. пункт 2). После ссылки на конкретизационную точку 4 добавляется
*	дополнительный элемент, не указывающий никуда (но требующийся для
*	корректной работы ряда функций).
*
*/

int main (int argc, char *argv []) {
	LITERO *kkh, *kks, *kps, *kph;
	
	setlocale (LC_ALL, "");
	textdomain ("refal");
	ARGC = argc;
	ARGV = argv;
	LLM = (LITERO *) malloc (sizeof (LITERO));
	LF = (FUNKCIO *) malloc (sizeof (FUNKCIO));
	if ((LLM == NULL) || (LF == NULL)) ref_eraro (1);
	LF->enk = 1;
	LF->dat.f = ref_cxesigi;
	LF->sekv = (FUNKCIO *) malloc (sizeof (FUNKCIO));
	if (LF->sekv == NULL) ref_eraro (1);
	(LF->sekv)->nom = (unsigned long *) malloc
		(sizeof (unsigned long) * NOM_LIM);
	if ((LF->sekv)->nom == NULL) ref_eraro (1);
	(LF->sekv)->enk = 0;
	(LF->sekv)->dat.s = NULL;
	(LF->sekv)->nom [0] = 36;
	(LF->sekv)->nom [1] = 115;
	(LF->sekv)->nom [2] = 116;
	(LF->sekv)->nom [3] = 97;
	(LF->sekv)->nom [4] = 114;
	(LF->sekv)->nom [5] = 116;
	(LF->sekv)->nom [6] = 0;
	(LF->sekv)->sekv = NULL;
	krei_literon (kkh);
	krei_literon (kks);
	krei_literon (kps);
	krei_literon (kph);
	krei_literon (kkh->ant);
	krei_literon (kph->sekv);
	ligi_literojn (kkh, kks);
	ligi_literojn (kks, kps);
	ligi_literojn (kps, kph);
	kkh->tip = KONK_K;
	kkh->dat.funk = LF;
	kks->tip = KONK_K;
	kks->dat.funk = LF->sekv;
	kps->tip = KONK_P;
	kps->dat.dual = kks;
	kph->tip = KONK_P;
	kph->dat.dual = kkh;
	krei_literon (LKP);
	krei_literon (LKP->sekv);
	krei_literon ((LKP->sekv)->sekv);
	(LKP->sekv)->ant = LKP;
	((LKP->sekv)->sekv)->ant = LKP->sekv;
	LKP->dat.dual = kps;
	(LKP->sekv)->dat.dual = kph;
	interp_krei_masxinon (argv [1]);
	ref_lancxi ();
	return 0;
}

